@@ -3,10 +3,10 @@
 
 This file is part of XEmacs.
 
-XEmacs is free software: you can redistribute it and/or modify it
+XEmacs is free software; you can redistribute it and/or modify it
 under the terms of the GNU General Public License as published by the
-Free Software Foundation, either version 3 of the License, or (at your
-option) any later version.
+Free Software Foundation; either version 2, or (at your option) any
+later version.
 
 XEmacs is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
@@ -14,22 +14,21 @@ FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 for more details.
 
 You should have received a copy of the GNU General Public License
-along with XEmacs.  If not, see <http://www.gnu.org/licenses/>.
+along with XEmacs; see the file COPYING.  If not, write to the Free
+Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
+02111-1307, USA.
+
    Geoff Voelker (voelker@cs.washington.edu) 7-29-94 */
 
 /* Adapted for XEmacs by David Hobley <david@spook-le0.cia.com.au> */
-/* Synced with FSF Emacs 19.34.6 by Marc Paquette <marcpa@cam.org>
-   (Note: Sync messages from Marc Paquette may indicate
-   incomplete synching, so beware.)
- */
-
-/* This file has been Mule-ized, Ben Wing, 4-13-02. */
+/* Synced with FSF Emacs 19.34.6 by Marc Paquette <marcpa@cam.org> */
 
 #include <config.h>
-#include "lisp.h"
+#include "lisp.h"  /* for VALMASK */
 
-#include "sysdep.h"
-#include "syswindows.h"
+#include <stdlib.h>
+
+#include "ntheap.h"
 
 /* This gives us the page size and the size of the allocation unit on NT.  */
 SYSTEM_INFO sysinfo_cache;
@@ -49,39 +48,39 @@ cache_system_info (void)
 }
 
 /* Round ADDRESS up to be aligned with ALIGN.  */
-URawbyte *
-round_to_next (URawbyte *address, unsigned long align)
+unsigned char *
+round_to_next (unsigned char *address, unsigned long align)
 {
   unsigned long tmp;
 
   tmp = (unsigned long) address;
   tmp = (tmp + align - 1) / align;
 
-  return (URawbyte *) (tmp * align);
+  return (unsigned char *) (tmp * align);
 }
 
 /* Info for keeping track of our heap.  */
-URawbyte *data_region_base = UNINIT_PTR;
-URawbyte *data_region_end = UNINIT_PTR;
-URawbyte *real_data_region_end = UNINIT_PTR;
-EMACS_UINT  data_region_size = UNINIT_EMACS_UINT;
-EMACS_UINT  reserved_heap_size = UNINIT_EMACS_UINT;
+unsigned char *data_region_base = UNINIT_PTR;
+unsigned char *data_region_end = UNINIT_PTR;
+unsigned char *real_data_region_end = UNINIT_PTR;
+unsigned long  data_region_size = UNINIT_LONG;
+unsigned long  reserved_heap_size = UNINIT_LONG;
 
 /* The start of the data segment.  */
-URawbyte *
+unsigned char *
 get_data_start (void)
 {
   return data_region_base;
 }
 
 /* The end of the data segment.  */
-URawbyte *
+unsigned char *
 get_data_end (void)
 {
   return data_region_end;
 }
 
-static URawbyte *
+static unsigned char *
 allocate_heap (void)
 {
   /* The base address for our GNU malloc heap is chosen in conjunction
@@ -120,9 +119,9 @@ allocate_heap (void)
      the region below the 256MB line for our malloc arena - 229MB is
      still a pretty decent arena to play in!  */
 
-  EMACS_UINT base = 0x01B00000;   /*  27MB */
+  unsigned long base = 0x01B00000;   /*  27MB */
   /* Temporary hack for the non-starting problem - use 28 (256Mb) rather than VALBITS (1Gb) */
-  EMACS_UINT end  = 1 << 28;      /* 256MB */
+  unsigned long end  = 1 << 28;      /* 256MB */
   void *ptr = NULL;
 
 #define NTHEAP_PROBE_BASE 1
@@ -145,7 +144,7 @@ allocate_heap (void)
 		      PAGE_NOACCESS);
 #endif
 
-  return (URawbyte *) ptr;
+  return (unsigned char*) ptr;
 }
 
 
@@ -174,7 +173,7 @@ sbrk (unsigned long increment)
   if (size < 0) 
     {
       int new_size;
-      URawbyte *new_data_region_end;
+      unsigned char *new_data_region_end;
 
       size = -size;
 
@@ -185,7 +184,7 @@ sbrk (unsigned long increment)
       /* We can only decommit full pages, so allow for 
 	 partial deallocation [cga].  */
       new_data_region_end = (data_region_end - size);
-      new_data_region_end = (URawbyte *)
+      new_data_region_end = (unsigned char *)
 	((long) (new_data_region_end + syspage_mask) & ~syspage_mask);
       new_size = real_data_region_end - new_data_region_end;
       real_data_region_end = new_data_region_end;
@@ -214,13 +213,70 @@ sbrk (unsigned long increment)
 
       /* We really only commit full pages, so record where
 	 the real end of committed memory is [cga].  */
-      real_data_region_end = (URawbyte *)
+      real_data_region_end = (unsigned char *)
 	  ((long) (data_region_end + syspage_mask) & ~syspage_mask);
     }
   
   return result;
 }
 
+#if !defined (CANNOT_DUMP) && !defined(HEAP_IN_DATA) && !defined(PDUMP)
+
+/* Recreate the heap from the data that was dumped to the executable.
+   EXECUTABLE_PATH tells us where to find the executable.  */
+void
+recreate_heap (char *executable_path)
+{
+  /* First reserve the upper part of our heap.  (We reserve first
+	 because there have been problems in the past where doing the
+	 mapping first has loaded DLLs into the VA space of our heap.)  */
+
+  /* Query the region at the end of the committed heap */
+  void *tmp;
+  MEMORY_BASIC_INFORMATION info;
+  DWORD size;
+  unsigned char* base = get_heap_end ();
+  unsigned char* end  = base + get_reserved_heap_size () - get_committed_heap_size ();
+  VirtualQuery (base, &info, sizeof info);
+  if (info.State != MEM_FREE)
+	{
+	  /* Oops, something has already reserved or commited it, nothing we can do but exit */
+	  char buf[256];
+	  wsprintf(buf,
+			   "XEmacs cannot start because the memory region required by the heap is not available.\n"
+			   "(BaseAddress = 0x%lx, AllocationBase = 0x%lx, Size = 0x%lx, State = %s, Type = %s)",
+			   info.BaseAddress, info.AllocationBase, info.RegionSize,
+			   info.State == MEM_COMMIT ? "COMMITED" : "RESERVED",
+			   info.Type == MEM_IMAGE ? "IMAGE" : info.Type == MEM_MAPPED ? "MAPPED" : "PRIVATE");
+	  MessageBox(NULL, buf, "XEmacs", MB_OK | MB_ICONSTOP);
+	  exit(1);
+	}
+
+  /* Now try and reserve as much as possible */
+  size = min (info.RegionSize, end - base);
+  tmp = VirtualAlloc (base, size, MEM_RESERVE, PAGE_NOACCESS);
+  if (!tmp)
+	{
+	  /* Can't reserve it, nothing we can do but exit */
+	  char buf[256];
+	  wsprintf(buf,
+			   "XEmacs cannot start because it couldn't reserve space required for the heap.\n"
+			   "(VirtualAlloc at 0x%lx of 0x%lx failed (%d))", base, size, GetLastError());
+	  MessageBox(NULL, buf, "XEmacs", MB_OK | MB_ICONSTOP);
+	  exit (1);
+	}
+
+  /* We read in the data for the .bss section from the executable
+     first and map in the heap from the executable second to prevent
+     any funny interactions between file I/O and file mapping.  */
+  read_in_bss (executable_path);
+  map_in_heap (executable_path);
+
+  /* Update system version information to match current system.  */
+  cache_system_info ();
+}
+#endif /* CANNOT_DUMP */
+
 /* Round the heap up to the given alignment.  */
 void
 round_heap (unsigned long align)
